#! /usr/bin/env python
"""
Fill is a script to fill the slices of a gcode file.

To run fill, install python 2.x on your machine, which is avaliable from http://www.python.org/download/

To use the preferences dialog you'll also need Tkinter, which probably came with the python installation.  If it did not, look for it at:
www.tcl.tk/software/tcltk/

To export a GNU Triangulated Surface file from Art of Illusion, you can use the Export GNU Triangulated Surface script at:
http://members.axion.net/~enrique/Export%20GNU%20Triangulated%20Surface.bsh

To bring it into Art of Illusion, drop it into the folder ArtOfIllusion/Scripts/Tools/.

The GNU Triangulated Surface format is supported by Mesh Viewer, and it is described at:
http://gts.sourceforge.net/reference/gts-surfaces.html#GTS-SURFACE-WRITE

To turn an STL file into filled gcode, first import the file using the STL import plugin in the import submenu of the file menu of Art of Illusion.
Then from the Scripts submenu in the Tools menu, choose Skeinforge and select the imported STL shape.  In the Fill radio button group,
choose Slice and click the 'Export Selected' checkbox.  Set the parameters and click OK.  Then type 'python fill.py' in a shell in the folder
which fill is in and when the dialog pops up, then click 'Fill', choose the file which you exported in
Export GNU Triangulated Surface and the filled file will be saved with the suffix '_fill'.

To write documentation for this program, open a shell in the fill.py directory, then type 'pydoc -w fill', then open 'fill.html' in a browser
or click on the '?' button in the dialog.  To use other functions of fill, type 'python' in a shell to run the python interpreter, then type 'import fill'
to import this program.

The computation intensive python modules will use psyco if it is available and run about twice as fast.  Psyco is described at:
http://psyco.sourceforge.net/index.html

The psyco download page is:
http://psyco.sourceforge.net/download.html

The following examples fillet the files Hollow Square.gcode & Hollow Square.gts.  The examples are run in a terminal in the folder which contains
Hollow Square.gcode, Hollow Square.gts and fill.py.  The preferences can be set in the dialog or by changing the preferences file 'fill.csv' with a text editor
or a spreadsheet program set to separate tabs.


> pydoc -w fill
wrote fill.html


> python fill.py
This brings up the dialog, after clicking 'Fill', the following is printed:
File Hollow Square.gts is being chain filled.
The filled file is saved as Hollow Square_fill.gcode


>python
Python 2.5.1 (r251:54863, Sep 22 2007, 01:43:31)
[GCC 4.2.1 (SUSE Linux)] on linux2
Type "help", "copyright", "credits" or "license" for more information.
>>> import fill
>>> fill.main()
File Hollow Square.gts is being filled.
The filled file is saved as Hollow Square_fill.gcode
It took 3 seconds to fill the file.


>>> fill.fillChainFile()
File Hollow Square.gts is being filled.
The filled file is saved as Hollow Square_fill.gcode
It took 3 seconds to fill the file.


>>> fill.fillFile()
File Hollow Square.gcode is being filled.
The filled file is saved as Hollow Square_fill.gcode
It took 3 seconds to fill the file.


>>> fill.getFillChainGcode("
( GCode generated by May 8, 2008 slice.py )
( Extruder Initialization )
..
many lines of gcode
..
")


>>> fill.getFillGcode("
( GCode generated by May 8, 2008 slice.py )
( Extruder Initialization )
..
many lines of gcode
..
")
"""

try:
	import psyco
	psyco.full()
except:
	pass
from vec3 import Vec3
import cStringIO
import euclidean
import gcodec
import intercircle
import math
import preferences
import slice
import time
import vectorwrite


__author__ = "Enrique Perez (perez_enrique@yahoo.com)"
__date__ = "$Date: 2008/28/04 $"
__license__ = "GPL 3.0"

#multiline perimeter, perpendicular span, one direction for while, split to weave, hex fill, loop inside sparse fill
#stretch for estimated contraction, tower
#much afterwards make multistep view, maybe simplify paths on top of loops actually I think we're already doing this
def addAroundClosest( paths, removedEndpointPoint ):
	"Add the closest removed endpoint to the path, with minimal twisting."
	closestDistanceSquared = 999999999.0
	closestPath = None
	closestPointIndex = None
	for path in paths:
		for pointIndex in range( len( path ) ):
			point = path[ pointIndex ]
			distanceSquared = point.distance2( removedEndpointPoint )
			if distanceSquared < closestDistanceSquared:
				closestDistanceSquared = distanceSquared
				closestPath = path
				closestPointIndex = pointIndex
	if closestPath == None:
		return
	if closestPointIndex + 1 == len( closestPath ):
		closestPath.insert( closestPointIndex + 1, removedEndpointPoint )
		return
	if closestPointIndex == 0:
		closestPath.insert( 0, removedEndpointPoint )
		return
	beforePoint = closestPath[ closestPointIndex - 1 ]
	centerPoint = closestPath[ closestPointIndex ]
	afterPoint = closestPath[ closestPointIndex + 1 ]
	beforeSegment = beforePoint.minus( removedEndpointPoint )
	beforeSegment.normalize()
	centerSegment = centerPoint.minus( removedEndpointPoint )
	centerSegment.normalize()
	afterSegment = afterPoint.minus( removedEndpointPoint )
	afterSegment.normalize()
	beforePlaneDot = euclidean.getPlaneDotPlusOne( beforeSegment, centerSegment )
	afterPlaneDot = euclidean.getPlaneDotPlusOne( centerSegment, afterSegment )
	if beforePlaneDot < afterPlaneDot:
		if beforePlaneDot < 2.0:
			closestPath.insert( closestPointIndex, removedEndpointPoint )
		return
	if afterPlaneDot < 2.0:
		closestPath.insert( closestPointIndex + 1, removedEndpointPoint )

def addPath( extrusionWidth, fill, path, rotationPlaneAngle ):
	"Add simplified path to fill."
	planeRotated = euclidean.getPathRoundZAxisByPlaneAngle( rotationPlaneAngle, euclidean.getSimplifiedLoop( path, extrusionWidth ) )
	fill.append( planeRotated )

def addSparseEndpoints( doubleExtrusionWidth, endpoints, fillDensity, fillLine, horizontalSegments, removedEndpoints, surroundingXIntersections ):
	"Add sparse endpoints."
	horizontalEndpoints = horizontalSegments[ fillLine ]
	for segment in horizontalEndpoints:
		addSparseEndpointsFromSegment( doubleExtrusionWidth, endpoints, fillDensity, fillLine, horizontalSegments, removedEndpoints, segment, surroundingXIntersections )

def addSparseEndpointsFromSegment( doubleExtrusionWidth, endpoints, fillDensity, fillLine, horizontalSegments, removedEndpoints, segment, surroundingXIntersections ):
	"Add sparse endpoints from a segment."
	endpointFirstPoint = segment[ 0 ].point
	endpointSecondPoint = segment[ 1 ].point
	shouldFill = fillLine < 1 or fillLine >= len( horizontalSegments ) - 1 or surroundingXIntersections == None
	if fillDensity > 0.0:
		if int( round( round( fillLine * fillDensity ) / fillDensity ) ) == fillLine:
			shouldFill = True
	if endpointFirstPoint.distance( endpointSecondPoint ) < doubleExtrusionWidth:
		shouldFill = True
	if shouldFill:
		endpoints += segment
		return
	if not isSegmentAround( horizontalSegments[ fillLine - 1 ], segment ):
		endpoints += segment
		return
	if not isSegmentAround( horizontalSegments[ fillLine + 1 ], segment ):
		endpoints += segment
		return
	for surroundingIndex in range( 0, len( surroundingXIntersections ), 2 ):
		surroundingXFirst = surroundingXIntersections[ surroundingIndex ]
		surroundingXSecond = surroundingXIntersections[ surroundingIndex + 1 ]
		if euclidean.isSegmentCompletelyInX( segment, surroundingXFirst, surroundingXSecond ):
			removedEndpoints += segment
			return
	endpoints += segment

def fillChainFile( filename = '' ):
	"Fill the slices of a gcode file.  Chain slice the file if it is a GNU TriangulatedSurface file.  If no filename is specified, fill the first unmodified gcode file in this folder."
	if filename == '':
		unmodified = gcodec.getGNUGcode()
		if len( unmodified ) == 0:
			print( "There are no unmodified gcode files in this folder." )
			return
		filename = unmodified[ 0 ]
	startTime = time.time()
	fillPreferences = FillPreferences()
	preferences.readPreferences( fillPreferences )
	print( 'File ' + gcodec.getSummarizedFilename( filename ) + ' is being chain filled.' )
	gcodeText = gcodec.getFileText( filename )
	if gcodeText == '':
		return
	suffixFilename = filename[ : filename.rfind( '.' ) ] + '_fill.gcode'
	gcodec.writeFileText( suffixFilename, getFillChainGcode( gcodeText, fillPreferences ) )
	print( 'The filled file is saved as ' + suffixFilename )
	if fillPreferences.writeSVG.value:
		vectorwrite.writeVectorFile( suffixFilename )
	print( 'It took ' + str( int( round( time.time() - startTime ) ) ) + ' seconds to fill the file.' )

def fillFile( filename = '' ):
	"Fill the slices of a gcode file.  If no filename is specified, fill the first unmodified gcode file in this folder."
	if filename == '':
		unmodified = gcodec.getUnmodifiedGCodeFiles()
		if len( unmodified ) == 0:
			print( "There are no unmodified gcode files in this folder." )
			return
		filename = unmodified[ 0 ]
	startTime = time.time()
	fillPreferences = FillPreferences()
	preferences.readPreferences( fillPreferences )
	print( 'File ' + gcodec.getSummarizedFilename( filename ) + ' is being filled.' )
	gcodeText = gcodec.getFileText( filename )
	if gcodeText == '':
		return
	suffixFilename = filename[ : filename.rfind( '.' ) ] + '_fill.gcode'
	gcodec.writeFileText( suffixFilename, getFillGcode( gcodeText, fillPreferences ) )
	print( 'The filled file is saved as ' + gcodec.getSummarizedFilename( suffixFilename ) )
	if fillPreferences.writeSVG.value:
		vectorwrite.writeVectorFile( suffixFilename )
	print( 'It took ' + str( int( round( time.time() - startTime ) ) ) + ' seconds to fill the file.' )

def getFillChainGcode( gcodeText, fillPreferences = None ):
	"Fill the slices of a gcode text.  Chain fill the gcode if it is not already sliced."
	if not gcodec.isProcedureDone( gcodeText, 'slice' ):
		gcodeText = slice.getSliceGcode( gcodeText )
	return getFillGcode( gcodeText, fillPreferences )

def getFillGcode( gcodeText, fillPreferences = None ):
	"Fill the slices of a gcode text."
	if gcodeText == '':
		return ''
	if gcodec.isProcedureDone( gcodeText, 'fill' ):
		return gcodeText
	if fillPreferences == None:
		fillPreferences = FillPreferences()
		preferences.readPreferences( fillPreferences )
	skein = FillSkein()
	skein.parseGcode( fillPreferences, gcodeText )
	return skein.output.getvalue()

def getHorizontalSegments( fillLoops, alreadyFilledArounds, y ):
	"Get horizontal segments inside loops."
	solidXIntersectionList = []
	euclidean.addXIntersectionsFromLoops( fillLoops, - 1, solidXIntersectionList, y )
	for alreadyFilledAroundIndex in range( len( alreadyFilledArounds ) ):
		alreadyFilledLoops = alreadyFilledArounds[ alreadyFilledAroundIndex ]
		for alreadyFilledLoopIndex in range( len( alreadyFilledLoops ) ):
			alreadyFilledLoop = alreadyFilledLoops[ alreadyFilledLoopIndex ]
			euclidean.addXIntersections( alreadyFilledLoop, alreadyFilledAroundIndex, solidXIntersectionList, y )
	return getSegmentsFromIntersections( solidXIntersectionList, y, fillLoops[ 0 ][ 0 ].z )

def getLoopStartingNearest( fromPoint, loop ):
	"Get loop starting at the nearest point to fromPoint."
	if fromPoint == None:
		return loop
	loop = loop[ : - 1 ]
	nearestIndex = int( round( euclidean.getNearestDistanceSquaredIndex( fromPoint, loop ).imag ) )
	aroundLoop = euclidean.getAroundLoop( nearestIndex, nearestIndex, loop )
	aroundLoop.append( aroundLoop[ 0 ] )
	return aroundLoop

def getSegmentsFromIntersections( solidXIntersectionList, y, z ):
	"Get endpoint segments from the intersections."
	segments = []
	xIntersectionList = []
	fill = False
	solid = False
	solidTable = {}
	solidXIntersectionList.sort( euclidean.compareSolidXByX )
	for solidX in solidXIntersectionList:
		solidXYInteger = int( solidX.imag )
		if solidXYInteger >= 0:
			toggleHashtable( solidTable, solidXYInteger, "" )
		else:
			fill = not fill
		oldSolid = solid
		solid = ( len( solidTable ) == 0 and fill )
		if oldSolid != solid:
			xIntersectionList.append( solidX.real )
	for xIntersectionIndex in range( 0, len( xIntersectionList ), 2 ):
		firstX = xIntersectionList[ xIntersectionIndex ]
		secondX = xIntersectionList[ xIntersectionIndex + 1 ]
		endpointFirst = Endpoint()
		endpointSecond = Endpoint().getFromOtherPoint( endpointFirst, Vec3().getFromXYZ( secondX, y, z ) )
		endpointFirst.getFromOtherPoint( endpointSecond, Vec3().getFromXYZ( firstX, y, z ) )
		segment = ( endpointFirst, endpointSecond )
		segments.append( segment )
	return segments

def getSurroundingXIntersections( alreadyFilledSize, doubleSolidSurfaceLayers, surroundingSlices, y ):
	"Get x intersections from surrounding layers."
	if len( surroundingSlices ) < doubleSolidSurfaceLayers:
		return None
	joinedX = []
	solidXIntersectionList = []
	for surroundingIndex in range( len( surroundingSlices ) ):
		surroundingSlice = surroundingSlices[ surroundingIndex ]
		euclidean.addXIntersectionsFromLoops( surroundingSlice, surroundingIndex, joinedX, y )
	solidTable = {}
	solid = False
	joinedX.sort( euclidean.compareSolidXByX )
	for solidX in joinedX:
		solidXYInteger = int( solidX.imag )
		toggleHashtable( solidTable, solidXYInteger, "" )
		oldSolid = solid
		solid = len( solidTable ) >= doubleSolidSurfaceLayers
		if oldSolid != solid:
			solidXIntersectionList.append( solidX.real )
	return solidXIntersectionList

def isSegmentAround( aroundSegments, segment ):
	"Determine if there is another segment around."
	for aroundSegment in aroundSegments:
		endpoint = aroundSegment[ 0 ]
		if isSegmentInX( segment, endpoint.point.x, endpoint.otherEndpoint.point.x ):
			return True
	return False

def isSegmentInX( segment, xFirst, xSecond ):
	"Determine if the segment overlaps within x."
	segmentFirstX = segment[ 0 ].point.x
	segmentSecondX = segment[ 1 ].point.x
	if min( segmentFirstX, segmentSecondX ) > max( xFirst, xSecond ):
		return False
	return max( segmentFirstX, segmentSecondX ) > min( xFirst, xSecond )

def toggleHashtable( hashtable, key, value ):
	"Toggle a hashtable between having and not having a key."
	if key in hashtable:
		del hashtable[ key ]
	else:
		hashtable[ key ] = value


class Endpoint:
	"The endpoint of a segment."
	def __repr__( self ):
		"Get the string representation of this Endpoint."
		return 'Endpoint ' + str( self.point ) + ' ' + str( self.otherEndpoint.point )

	def getFromOtherPoint( self, otherEndpoint, point ):
		"Initialize from other endpoint."
		self.otherEndpoint = otherEndpoint
		self.point = point
		return self

	def getNearestEndpoint( self, endpoints ):
		"Get nearest endpoint."
		smallestDistanceSquared = 999999999999999999.0
		nearestEndpoint = None
		for endpoint in endpoints:
			distanceSquared = self.point.distance2( endpoint.point )
			if distanceSquared < smallestDistanceSquared:
				smallestDistanceSquared = distanceSquared
				nearestEndpoint = endpoint
		return nearestEndpoint

	def getNearestMiss( self, arounds, endpoints, extrusionWidth, stretchedXSegments ):
		"Get the nearest endpoint which the segment to that enpoint misses the other extrusions."
		smallestDistanceSquared = 999999999999999999.0
		nearestMiss = None
		for endpoint in endpoints:
			segment = endpoint.point.minus( self.point )
			segmentYMirror = complex( segment.x, - segment.y )
			segmentYMirror *= 1.0 / abs( segmentYMirror )
			segmentFirstPoint = euclidean.getRoundZAxisByPlaneAngle( segmentYMirror, self.point )
			segmentSecondPoint = euclidean.getRoundZAxisByPlaneAngle( segmentYMirror, endpoint.point )
			distanceSquared = self.point.distance2( endpoint.point )
			if distanceSquared < smallestDistanceSquared:
				if not self.isLoopListCrossingInsideXSegment( arounds, segmentFirstPoint.x, segmentSecondPoint.x, segmentYMirror, segmentFirstPoint.y ):
					if not self.isPointCrossingSegments( extrusionWidth, endpoint.point, stretchedXSegments ):
						smallestDistanceSquared = distanceSquared
						nearestMiss = endpoint
		return nearestMiss

	def isLoopListCrossingInsideXSegment( self, loopList, segmentFirstX, segmentSecondX, segmentYMirror, y ):
		"Determine if the loop list is crossing inside the x segment."
		for alreadyFilledLoop in loopList:
			loop = euclidean.getPathRoundZAxisByPlaneAngle( segmentYMirror, alreadyFilledLoop )
			for pointIndex in range( len( loop ) ):
				pointFirst = loop[ pointIndex ]
				pointSecond = loop[ ( pointIndex + 1 ) % len( loop ) ]
				if euclidean.isLineCrossingInsideXSegment( segmentFirstX, segmentSecondX, pointFirst, pointSecond, y ):
					return True
		return False

	def isPointCrossingSegments( self, extrusionWidth, inputPoint, stretchedXSegments ):
		"Determine if the segment to the point is crossing a stretched x segment."
		endpointRadius = 0.1 * extrusionWidth
		segment = inputPoint.minus( self.point )
		alongSegmentLength = 0.333 * segment.length()
		segment.scale( alongSegmentLength / segment.length() )
		attractedPoint = self.point.plus( segment )
		attractedInputPoint = inputPoint.minus( segment )
		for stretchedXSegment in stretchedXSegments:
			if euclidean.isLineCrossingInsideXSegment( stretchedXSegment.xMinimum, stretchedXSegment.xMaximum, attractedPoint, attractedInputPoint, stretchedXSegment.y ):
				return True
		return False


class FillSkein:
	"A class to fill a skein of extrusions."
	def __init__( self ):
		self.extruderActive = False
		self.filledLayers = []
		self.lastOutputPoint = None
		self.layer = None
		self.layers = []
		self.lineIndex = 0
		self.oldLocation = None
		self.output = cStringIO.StringIO()
		self.thread = None

	def addFill( self, layerIndex ):
		"Add fill to the slice layer."
		alreadyFilledArounds = []
		arounds = []
		back = - 999999999.0
		doubleExtrusionWidth = 2.0 * self.extrusionWidth
		endpoints = []
		fill = []
		fillEdgeIndex = 0
		aroundInset = self.fillInset - 0.3 * self.halfExtrusionWidth
		front = 999999999.0
		layer = self.layers[ layerIndex ]
		filledLayer = layer[ : ]
		self.filledLayers.append( filledLayer )
		oneAndAHalfExtrusionWidth = 1.5 * self.extrusionWidth
		layerRotationAroundZAngle = self.getLayerRotationAroundZAngle( layerIndex )
		reverseRotationAroundZAngle = complex( layerRotationAroundZAngle.real, - layerRotationAroundZAngle.imag )
		rotatedExtruderLoops = []
		stretch = 0.1 * self.extrusionWidth
		for thread in layer:
			alreadyFilledLoop = []
			alreadyFilledArounds.append( alreadyFilledLoop )
			loop = thread[ 1 : ]
			planeRotatedPerimeter = euclidean.getPathRoundZAxisByPlaneAngle( reverseRotationAroundZAngle, loop )
			alreadyFilledLoop.append( planeRotatedPerimeter )
			centers = intercircle.getCentersfromLoopDirection( euclidean.isWiddershins( planeRotatedPerimeter ), planeRotatedPerimeter, oneAndAHalfExtrusionWidth )
			for center in centers:
				inset = intercircle.getInsetFromClockwiseLoop( center, self.fillInset )
				if euclidean.isWiddershins( center ) == euclidean.isWiddershins( inset ):
					around = intercircle.getInsetFromClockwiseLoop( center, aroundInset )
					rotatedExtruderLoops.append( intercircle.getInsetFromClockwiseLoop( center, self.halfFillInset ) )
					alreadyFilledLoop.append( inset )
					arounds.append( around )
					for point in inset:
						back = max( back, point.y )
						front = min( front, point.y )
		fillWidth = back - front
		numberOfIntervals = int( math.floor( fillWidth / self.extrusionWidth ) )
		fillRemainder = fillWidth - float( numberOfIntervals ) * self.extrusionWidth
		halfFillRemainder = 0.5 * fillRemainder
		back -= halfFillRemainder
		front += halfFillRemainder
		surroundingSlices = []
		layerRemainder = layerIndex % int( round( self.fillPreferences.diaphragmPeriod.value ) )
		if layerRemainder >= int( round( self.fillPreferences.diaphragmThickness.value ) ):
			for surroundingIndex in range( 1, self.solidSurfaceLayers + 1 ):
				self.addRotatedSlice( layerIndex - surroundingIndex, reverseRotationAroundZAngle, surroundingSlices )
				self.addRotatedSlice( layerIndex + surroundingIndex, reverseRotationAroundZAngle, surroundingSlices )
		self.infillType = 'sparse'
		if len( surroundingSlices ) < self.doubleSolidSurfaceLayers:
			self.infillType = 'solid'
		horizontalSegments = []
		for fillLine in range( numberOfIntervals + 1 ):
			y = front + float( fillLine ) * self.extrusionWidth
			lineSegments = getHorizontalSegments( rotatedExtruderLoops, alreadyFilledArounds, y )
			horizontalSegments.append( lineSegments )
		removedEndpoints = []
		for fillLine in range( len( horizontalSegments ) ):
			y = front + float( fillLine ) * self.extrusionWidth
			horizontalEndpoints = horizontalSegments[ fillLine ]
			surroundingXIntersections = getSurroundingXIntersections( len( alreadyFilledArounds ), self.doubleSolidSurfaceLayers, surroundingSlices, y )
			addSparseEndpoints( doubleExtrusionWidth, endpoints, self.fillDensity, fillLine, horizontalSegments, removedEndpoints, surroundingXIntersections )
		if len( endpoints ) < 1:
			return
		stretchedXSegments = []
		for beginningEndpoint in endpoints[ : : 2 ]:
			beginningPoint = beginningEndpoint.point
			stretchedXSegment = StretchedXSegment().getFromXYStretch( beginningPoint.x, beginningPoint.y, beginningEndpoint.otherEndpoint.point.x, stretch )
			stretchedXSegments.append( stretchedXSegment )
		endpointFirst = endpoints[ 0 ]
		endpoints.remove( endpointFirst )
		otherEndpoint = endpointFirst.otherEndpoint
		endpoints.remove( otherEndpoint )
		path = []
		paths = []
		if len( endpoints ) > 1:
			nextEndpoint = otherEndpoint.getNearestMiss( arounds, endpoints, self.extrusionWidth, stretchedXSegments )
			if nextEndpoint.point.distance2( endpointFirst.point ) < nextEndpoint.point.distance2( otherEndpoint.point ):
				endpointFirst = endpointFirst.otherEndpoint
				otherEndpoint = endpointFirst.otherEndpoint
		path.append( endpointFirst.point )
		path.append( otherEndpoint.point )
		while len( endpoints ) > 1:
			nextEndpoint = otherEndpoint.getNearestMiss( arounds, endpoints, self.extrusionWidth, stretchedXSegments )
			if nextEndpoint == None:
				paths.append( path )
				path = []
				nextEndpoint = otherEndpoint.getNearestEndpoint( endpoints )
			path.append( nextEndpoint.point )
			endpoints.remove( nextEndpoint )
			otherEndpoint = nextEndpoint.otherEndpoint
			path.append( otherEndpoint.point )
			endpoints.remove( otherEndpoint )
		paths.append( path )
		for removedEndpoint in removedEndpoints:
			addAroundClosest( paths, removedEndpoint.point )
		for path in paths:
			addPath( self.extrusionWidth, fill, path, layerRotationAroundZAngle )
		filledLayer += fill

	def addFilledLayerToOutput( self, layerIndex ):
		"Add a filled layer to the output."
		filledLayer = self.filledLayers[ layerIndex ]
		self.addLine( '( Extruder paths for layer ' + str( layerIndex ) + ' )' ) # GCode formatted comment
		self.addLine( 'M113 (' + str( layerIndex ) ) # Indicate that a new layer is starting.
		for thread in filledLayer:
			self.addGcodeFromThread( thread )

	def addGcodeMovement( self, point ):
		"Add a movement to the output."
		self.lastOutputPoint = point
		self.output.write( "G1 X" + euclidean.getRoundedToThreePlaces( point.x ) + " Y" + euclidean.getRoundedToThreePlaces( point.y ) )
		self.addLine( " Z" + euclidean.getRoundedToThreePlaces( point.z ) + " F" + euclidean.getRoundedToThreePlaces( self.feedratePerMinute ) )

	def addGcodeFromThread( self, thread ):
		"Add a thread to the output."
		isLoop = False
		if len( thread ) > 3:
			if thread[ 0 ].equals( thread[ - 1 ] ):
				isLoop = True
				thread = getLoopStartingNearest( self.lastOutputPoint, thread )
		if len( thread ) > 0:
			self.addGcodeMovement( thread[ 0 ] )
		else:
			print( "zero length vertex positions array which was skipped over, this should never happen" )
		if len( thread ) < 2:
			return
		self.addLine( "M101" ) # Turn extruder on.
		if isLoop:
			self.addLine( 'M114' ) # Indicate that a loop is beginning.
		for point in thread[ 1 : ]:
			self.addGcodeMovement( point )
		self.addLine( "M103" ) # Turn extruder off.

	def addLine( self, line ):
		"Add a line of text and a newline to the output."
		self.output.write( line + "\n" )

	def addRotatedSlice( self, layerIndex, reverseRotationAroundZAngle, surroundingSlices ):
		"Add a rotated slice to the surrounding slices."
		if layerIndex < 0 or layerIndex >= len( self.layers ):
			return
		layer = self.layers[ layerIndex ]
		rotatedSlice = []
		for thread in layer:
			planeRotatedLoop = euclidean.getPathRoundZAxisByPlaneAngle( reverseRotationAroundZAngle, thread[ 1 : ] )
			rotatedSlice.append( planeRotatedLoop )
		surroundingSlices.append( rotatedSlice )

	def addShutdownToOutput( self ):
		"Add shutdown gcode to the output."
		self.addLine( '( Extruder Shut Down )' ) # GCode formatted comment
		self.addLine( "M103" ) # Turn extruder off.

	def addToThread( self, location ):
		"Add a location to thread."
		if self.oldLocation == None:
			return
		if self.layer == None:
			self.layer = []
			self.layers.append( self.layer )
		if self.thread == None:
			self.thread = [ self.oldLocation ]
			self.layer.append( self.thread )
		self.thread.append( location )

	def getLayerRotationAngle( self, layerIndex ):
		"Get the angle the layer is rotated by."
		return self.fillBeginRotation + float( ( layerIndex % 2 ) * self.fillOddLayerExtraRotation )

	def getLayerRotationAroundZAngle( self, layerIndex ):
		"Get the around z angle of the amount the layer is rotated by."
		return euclidean.getPolar( self.getLayerRotationAngle( layerIndex ), 1.0 )

	def linearMove( self, splitLine ):
		"Add a linear move to the thread."
		location = gcodec.getLocationFromSplitLine( self.oldLocation, splitLine )
		if self.extruderActive:
			self.addToThread( location )
		self.oldLocation = location

	def parseGcode( self, fillPreferences, gcodeText ):
		"Parse gcode text and store the bevel gcode."
		self.fillPreferences = fillPreferences
		self.lines = gcodec.getTextLines( gcodeText )
		self.parseInitialization()
		self.feedratePerMinute = 60.0 * fillPreferences.feedratePerSecond.value
		self.fillDensity = fillPreferences.fillDensity.value
		self.fillBeginRotation = math.radians( fillPreferences.fillBeginRotation.value )
		self.fillOddLayerExtraRotation = math.radians( fillPreferences.fillOddLayerExtraRotation.value )
		self.solidSurfaceLayers = int( round( self.fillPreferences.solidSurfaceLayers.value ) )
		self.doubleSolidSurfaceLayers = self.solidSurfaceLayers + self.solidSurfaceLayers
		for line in self.lines[ self.lineIndex : ]:
			self.parseLine( line )
		for layerIndex in range( len( self.layers ) ):
			self.addFill( layerIndex )
			self.addFilledLayerToOutput( layerIndex )
		self.addShutdownToOutput()

	def parseInitialization( self ):
		"Parse gcode initialization and store the parameters."
		for self.lineIndex in range( len( self.lines ) ):
			line = self.lines[ self.lineIndex ]
			splitLine = line.split( ' ' )
			firstWord = ''
			if len( splitLine ) > 0:
				firstWord = splitLine[ 0 ]
			if firstWord == 'M109':
				self.extrusionWidth = gcodec.getDoubleAfterFirstLetter( splitLine[ 1 ] )
#				self.extrusionWidth *= 0.5
				self.halfExtrusionWidth = 0.5 * self.extrusionWidth
				self.absoluteFillPerimeterOverlap = self.halfExtrusionWidth * ( 1.0 - self.fillPreferences.infillPerimeterOverlap.value )
				self.fillInset = self.halfExtrusionWidth + self.absoluteFillPerimeterOverlap
				self.halfFillInset = 0.5 * self.fillInset
				self.addLine( 'M115 (' + str( self.halfFillInset ) ) # Set half fill inset.
			elif firstWord == 'M112':
				self.addLine( 'M111 (fill)' )
				self.addLine( line )
				return
			self.addLine( line )

	def parseLine( self, line ):
		"Parse a gcode line and add it to the fill skein."
		splitLine = line.split( ' ' )
		if len( splitLine ) < 1:
			return
		firstWord = splitLine[ 0 ]
		if firstWord == 'G1':
			self.linearMove( splitLine )
		elif firstWord == 'M101':
			self.extruderActive = True
		elif firstWord == 'M103':
			self.extruderActive = False
			self.thread = None
		elif firstWord == 'M113':
			self.layer = None
			self.thread = None


class FillPreferences:
	"A class to handle the fill preferences."
	def __init__( self ):
		"Set the default preferences, execute title & preferences filename."
		#Set the default preferences.
		self.diaphragmPeriod = preferences.IntPreference().getFromValue( 'Diaphragm Period (layers):', 9.0 )
		self.diaphragmThickness = preferences.IntPreference().getFromValue( 'Diaphragm Thickness (layers):', 2.0 )
		self.feedratePerSecond = preferences.FloatPreference().getFromValue( 'Feedrate (mm/s):', 10.0 )
		self.filenameInput = preferences.Filename().getFromFilename( [ ( 'GNU Triangulated Surface text files', '*.gts' ), ( 'Gcode text files', '*.gcode' ) ], 'Open File to be Filled', '' )
		self.fillBeginRotation = preferences.FloatPreference().getFromValue( 'Fill Begin Rotation (degrees):', 45.0 )
		self.fillDensity = preferences.FloatPreference().getFromValue( 'Fill Density (ratio):', 0.5 )
		self.fillOddLayerExtraRotation = preferences.FloatPreference().getFromValue( 'Fill Odd Layer Extra Rotation (degrees):', 90.0 )
		self.infillPerimeterOverlap = preferences.FloatPreference().getFromValue( 'Infill Perimeter Overlap (ratio):', 0.1 )
		self.solidSurfaceLayers = preferences.IntPreference().getFromValue( 'Solid Surface Layers (integer):', 2.0 )
		self.writeSVG = preferences.BooleanPreference().getFromValue( 'Write Scalable Vector Graphics:', True )
		directoryRadio = []
		self.directoryPreference = preferences.RadioLabel().getFromRadioLabel( 'Fill All Unmodified Files in a Directory', 'File or Directory Choice:', directoryRadio, False )
		self.filePreference = preferences.Radio().getFromRadio( 'Fill File', directoryRadio, True )
		#Create the archive, title of the execute button, title of the dialog & preferences filename.
		self.archive = [
			self.diaphragmPeriod,
			self.diaphragmThickness,
			self.feedratePerSecond,
			self.filenameInput,
			self.fillBeginRotation,
			self.fillDensity,
			self.fillOddLayerExtraRotation,
			self.infillPerimeterOverlap,
			self.solidSurfaceLayers,
			self.writeSVG,
			self.directoryPreference,
			self.filePreference ]
		self.executeTitle = 'Fill'
#		self.filename = getPreferencesFilePath( 'fill.csv' )
		self.filenamePreferences = 'fill.csv'
		self.filenameHelp = 'fill.html'
		self.title = 'Fill Preferences'

	def execute( self ):
		"Fill button has been clicked."
		filenames = gcodec.getGcodeDirectoryOrFile( self.directoryPreference.value, self.filenameInput.value, self.filenameInput.wasCancelled )
		for filename in filenames:
			fillChainFile( filename )


class StretchedXSegment:
	"A stretched x segment."
	def __repr__( self ):
		"Get the string representation of this StretchedXSegment."
		return str( self.xMinimum ) + ' ' + str( self.xMaximum ) + ' ' + str( self.y )

	def getFromXYStretch( self, firstX, y, secondX, stretch ):
		"Initialize from x, y, and stretch."
		self.xMaximum = max( firstX, secondX ) + stretch
		self.xMinimum = min( firstX, secondX ) - stretch
		self.y = y
		return self


def main( hashtable = None ):
	"Display the fill dialog."
	preferences.displayDialog( FillPreferences() )

if __name__ == "__main__":
	main()
